﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using HeuristicLab.Clients.Hive.CloudManager.Azure;
using HeuristicLab.Clients.Hive.CloudManager.Model;
using HeuristicLab.Core;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;

namespace HeuristicLab.Clients.Hive.CloudManager.Views {
  public partial class AddAzureServiceDialog : Form {
    private BindingList<Subscription> subscriptions;
    private BindingList<HostedService> hostedServices;
    private BindingList<string> locations;
    private BindingList<AffinityGroup> affinityGroups;
    private BindingList<StorageService> storageServices;

    private BackgroundWorker workerUpdate;
    private BackgroundWorker workerCreate;

    private bool isInitializing;
    private bool bwCompleted;
    private bool closePending;

    private string certificateFile;
    private string certificatePassword;

    public AddAzureServiceDialog(IItemList<Subscription> subs) {
      InitializeComponent();
      isInitializing = true;
      bwCompleted = true;

      this.subscriptions = new BindingList<Subscription>(subs);
      this.hostedServices = new BindingList<HostedService>();
      this.locations = new BindingList<string>();
      this.affinityGroups = new BindingList<AffinityGroup>();
      this.storageServices = new BindingList<StorageService>();

      SetLookupBinding(cmbLocation, locations, "", "");
      SetLookupBinding(cmbAffinityGroup, affinityGroups, "Label", "Name");
      SetLookupBinding(cmbChooseHostedService, hostedServices, "ServiceName", "ServiceName");
      SetLookupBinding(cmbChooseSubscription, subscriptions, "SubscriptionName", "SubscriptionID");
      SetLookupBinding(cmbStorageServices, storageServices, "ServiceName", "Url");

      if (!cbNewHostedService.Checked) {
        SetGroupBoxControlsEnabled(gbNewHostedService, cbNewHostedService.Checked);
      }

      certificateFile = string.Empty;
      certificatePassword = string.Empty;

      workerUpdate = new BackgroundWorker();
      workerUpdate.DoWork += new DoWorkEventHandler(UpdateTask);
      workerUpdate.RunWorkerCompleted += new RunWorkerCompletedEventHandler(WorkerCompleted);

      workerCreate = new BackgroundWorker();
      workerCreate.DoWork += new DoWorkEventHandler(CreateDeploymentTask);
      workerCreate.RunWorkerCompleted += new RunWorkerCompletedEventHandler(CreateDeploymentCompleted);

      isInitializing = false;
      cmbChooseSubscription_SelectedValueChanged(this, EventArgs.Empty);
    }

    protected override void OnFormClosing(FormClosingEventArgs e) {
      if (!bwCompleted) {
        SetAllControlsEnabled(this, false);
        e.Cancel = true;
        closePending = true;
      }
    }

    private void SetLookupBinding(ListControl toBind, object dataSource, string displayMember, string valueMember) {
      toBind.DisplayMember = displayMember;
      toBind.ValueMember = valueMember;
      toBind.DataSource = dataSource;
    }

    private void SetGroupBoxControlsEnabled(GroupBox gb, bool isEnabled) {
      foreach (Control ctrl in gb.Controls) {
        ctrl.Enabled = isEnabled;
      }
    }

    private void SetAllControlsEnabled(Form frm, bool isEnabled) {
      foreach (Control ctrl in frm.Controls) {
        if (!(ctrl is ProgressBar)) {
          ctrl.Enabled = isEnabled;
        }
      }
    }

    private void SetBindingList<T>(ref BindingList<T> bindinglist, List<T> list) {
      if (list != null) {
        BindingList<T> newList = new BindingList<T>(list);
        bindinglist = newList;
      }
    }

    private void UpdateBindingList<T>(BindingList<T> bindinglist, List<T> list) {
      if (list != null) {
        bindinglist.Clear();
        foreach (T item in list)
          bindinglist.Add(item);
      }
    }

    private void UpdateSubscriptionItem(BindingList<Subscription> bindinglist, Subscription item) {
      if (item != null) {
        foreach (Subscription blItem in bindinglist)
          if (blItem.Equals(item)) {
            blItem.Merge(item);
          }
      }
    }

    private void cbNewHostedService_CheckedChanged(object sender, EventArgs e) {
      if (cbNewHostedService.Checked) {
        Subscription subscription = (Subscription)cmbChooseSubscription.SelectedItem;
        if (subscription.CurrentHostedServices == subscription.MaxHostedServices) {
          MessageBox.Show("You have already created the maximum number of hosted services", "Limit Reached", MessageBoxButtons.OK, MessageBoxIcon.Error);
          cbNewHostedService.Checked = false;
        } else {
          SetGroupBoxControlsEnabled(gbNewHostedService, cbNewHostedService.Checked);
          cmbChooseHostedService.Enabled = false;
        }
      } else {
        // checked == false
        SetGroupBoxControlsEnabled(gbNewHostedService, false);
        if (hostedServices.Count == 0) {
          cbNewHostedService.Checked = true;
        }
      }
    }

    private void cmbChooseSubscription_SelectedValueChanged(object sender, EventArgs e) {
      if (!isInitializing) {
        Subscription item = (Subscription)cmbChooseSubscription.SelectedItem;
        if (item != null) {
          bwCompleted = false;
          progressBar.Visible = true;
          SetAllControlsEnabled(this, false);
          workerUpdate.RunWorkerAsync(item);
        }
      }
    }

    private void btnAddCertificate_Click(object sender, EventArgs e) {
      using (var form = new AddCertificate()) {
        form.ShowDialog();
        if (!form.ErrorOccured) {
          certificateFile = form.CertificateFile;
          certificatePassword = form.CertificatePassword;
          lblCertificateFile.Text = certificateFile;
        }
      }
    }

    private void btnCancel_Click(object sender, EventArgs e) {
      this.Close();
    }

    private void btnOk_Click(object sender, EventArgs e) {
      if (cmbChooseSubscription.SelectedItem != null) {
        Subscription sub = (Subscription)cmbChooseSubscription.SelectedItem;
        // the controls collection can be the whole form or just a panel or group
        if (Validation.hasValidationErrors(this.Controls)) {
          return;
        } else {
          bool newHostedServiceChecked = cbNewHostedService.Checked;
          bool regionChecked = rbRegion.Checked;
          string certFile = certificateFile;
          string certPw = certificatePassword;
          int instanceCount = Convert.ToInt32(tbInstanceCount.Text);
          HostedService hostedService = new HostedService();
          hostedService.ServiceName = tbServiceName.Text;
          HostedServiceProperties properties = new HostedServiceProperties();
          if (regionChecked) {
            properties.AffinityGroup = string.Empty;
            properties.Location = (string)cmbLocation.SelectedItem;
          } else {
            properties.AffinityGroup = ((AffinityGroup)cmbAffinityGroup.SelectedItem).Name;
            properties.Location = string.Empty;
          }

          if (cmbAffinityGroup.SelectedItem != null) {
            properties.AffinityGroup = ((AffinityGroup)cmbAffinityGroup.SelectedItem).Name;
          }
          properties.Label = tbLabel.Text;
          properties.Location = (string)cmbLocation.SelectedItem;
          hostedService.HostedServiceProperties = properties;

          //var parameters1 = Tuple.Create<Subscription, bool, bool, string, string, HostedService, int>
          //  (sub, newHostedServiceChecked, regionChecked, certificateFile, certificatePassword, hostedService, instanceCount);
          CreateDeploymentData parameters = new CreateDeploymentData();
          parameters.Subscription = sub;
          parameters.CreateNewHosteService = newHostedServiceChecked;
          parameters.UseRegion = regionChecked;
          parameters.CertificateFilePath = certificateFile;
          parameters.CertificatePassword = certificatePassword;
          parameters.HostedService = hostedService;
          parameters.InstanceCount = instanceCount;
          parameters.StorageService = (StorageService)cmbStorageServices.SelectedItem;
          parameters.BlobContainerName = tbBlobContainer.Text;

          bwCompleted = false;
          progressBar.Visible = true;
          SetAllControlsEnabled(this, false);
          workerCreate.RunWorkerAsync(parameters);
        }
      }
    }

    private void UpdateTask(object sender, DoWorkEventArgs e) {
      try {
        Subscription item = (Subscription)e.Argument;
        item = CloudManagerClient.Instance.AzureProvider.GetSubscriptionInfo(item.SubscriptionID, item.CertificateThumbprint);
        this.Invoke((MethodInvoker)delegate { lblCores.Text = item.CurrentCoreCount + " / " + item.MaxCoreCount; });

        List<HostedService> services = CloudManagerClient.Instance.AzureProvider.ListHostedServices(item);
        List<string> locs = CloudManagerClient.Instance.AzureProvider.ListLocations(item);
        List<AffinityGroup> groups = CloudManagerClient.Instance.AzureProvider.ListAffinityGroups(item);
        List<StorageService> storage = CloudManagerClient.Instance.AzureProvider.ListStorageServices(item);

        this.Invoke((MethodInvoker)delegate { UpdateSubscriptionItem(subscriptions, item); });
        this.Invoke((MethodInvoker)delegate { UpdateBindingList<HostedService>(hostedServices, services); });
        this.Invoke((MethodInvoker)delegate { UpdateBindingList<string>(locations, locs); });
        this.Invoke((MethodInvoker)delegate { UpdateBindingList<AffinityGroup>(affinityGroups, groups); });
        this.Invoke((MethodInvoker)delegate { UpdateBindingList<StorageService>(storageServices, storage); });
      }
      catch (Exception ex) {
        MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
      }
    }

    private void WorkerCompleted(object sender, RunWorkerCompletedEventArgs e) {
      SetAllControlsEnabled(this, true);
      progressBar.Visible = false;
      bwCompleted = true;
      if (closePending) {
        this.Close();
      }
    }

    private void CreateDeploymentTask(object sender, DoWorkEventArgs e) {
      bool errorOccured = false;
      CreateDeploymentData parameters = (CreateDeploymentData)e.Argument;

      Subscription sub = parameters.Subscription;
      bool newHostedServiceChecked = parameters.CreateNewHosteService;
      bool regionChecked = parameters.UseRegion;
      string certFile = parameters.CertificateFilePath;
      string certPw = parameters.CertificatePassword;
      HostedService hostedService = parameters.HostedService;
      int instanceCount = parameters.InstanceCount;
      StorageService storageService = parameters.StorageService;
      string blobContainerName = parameters.BlobContainerName;
      bool createContainerIfNotExists = parameters.CreateContainerIfNotExists;

      CloudStorageAccount storageAccount;
      CloudBlobClient blobClient;
      CloudBlobContainer blobContainer;

      // STEP 1 - Initialize storage account
      try {
        // get access key to storage service
        StorageServiceKeys keys = CloudManagerClient.Instance.AzureProvider.GetStorageKeys(sub, storageService.ServiceName);
        storageAccount = CloudStorageAccount.Parse(string.Format(Constants.StorageServiceConnectionFormat, storageService.ServiceName, keys.Primary));
        blobClient = storageAccount.CreateCloudBlobClient();
        blobContainer = blobClient.GetContainerReference(blobContainerName);
        if (createContainerIfNotExists) {
          blobContainer.CreateIfNotExist();
        } else {
          if (!blobContainer.Exists()) {
            throw new ApplicationException("Specified blob container doesn't exist.");
          }
        }
      }
      catch (Exception ex) {
        errorOccured = true;
        MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
      }


      // STEP 1 - Create Hosted Service
      try {
        if ((!errorOccured) && (newHostedServiceChecked)) {
          if (regionChecked) {
            CloudManagerClient.Instance.AzureProvider.CreateHostedService(sub, hostedService.ServiceName, hostedService.HostedServiceProperties.Label, hostedService.HostedServiceProperties.Description, hostedService.HostedServiceProperties.Location);
          } else {
            CloudManagerClient.Instance.AzureProvider.CreateHostedService(sub, hostedService.ServiceName, hostedService.HostedServiceProperties.Label, hostedService.HostedServiceProperties.Description, new AffinityGroup() { Name = hostedService.HostedServiceProperties.AffinityGroup });
          }
        }
      }
      catch (Exception ex) {
        errorOccured = true;
        MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
      }

      // STEP 2 - Add Certificate
      try {
        if ((!errorOccured) && (certFile != string.Empty)) {
          CloudManagerClient.Instance.AzureProvider.AddCertificate(sub, hostedService, certFile, certPw);
        }
      }
      catch (Exception ex) {
        errorOccured = true;
        MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
        if (newHostedServiceChecked) {
          CloudManagerClient.Instance.AzureProvider.DeleteHostedService(sub, hostedService.ServiceName);
        }

      }

      // STEP 3 - Create Deployment
      try {
        if (!errorOccured) {
          CloudManagerClient.Instance.AzureProvider.CreateDeployment(sub, hostedService.ServiceName, Guid.NewGuid().ToString(), Constants.DeploymentSlotStaging, Constants.DeploymentPackageUrl, Constants.DeploymentConfigurationUrl, Constants.DeploymentLabel, instanceCount);
        }
      }
      catch (Exception ex) {
        errorOccured = true;
        MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
        if (newHostedServiceChecked) {
          CloudManagerClient.Instance.AzureProvider.DeleteHostedService(sub, hostedService.ServiceName);
        }
      }

      e.Result = errorOccured;
    }

    private void CreateDeploymentCompleted(object sender, RunWorkerCompletedEventArgs e) {
      SetAllControlsEnabled(this, true);
      progressBar.Visible = false;
      bwCompleted = true;
      bool errorOccured = (bool)e.Result;
      if (closePending) {
        this.Close();
      } else if (!errorOccured) {
        this.Close();
      } else {
        this.Show();
      }
    }

    private void tbServiceName_Validating(object sender, CancelEventArgs e) {
      if (cbNewHostedService.Checked) {
        string url = tbServiceName.Text.Trim();
        if (url == String.Empty) {
          errorProvider.SetError(tbServiceName, "A URL is required for the hosted service.");
          e.Cancel = true;
        } else if (!isValidHostedServiceUrl(url)) {
          errorProvider.SetError(tbServiceName, "A hosted service name must be between 1 and 63 " +
                                                "characters long, and be composed of letters, numbers " +
                                                "and hyphens. Hyphens shouldn't be first or last letter.");
          e.Cancel = true;
        } else {
          errorProvider.SetError(tbServiceName, "");
        }
      }
    }

    private void tbLabel_Validating(object sender, CancelEventArgs e) {
      if (cbNewHostedService.Checked) {
        if (tbLabel.Text.Trim() == String.Empty) {
          errorProvider.SetError(tbLabel, "A name is required for the hosted service.");
          e.Cancel = true;
        } else {
          errorProvider.SetError(tbLabel, "");
        }
      }
    }

    private bool isValidHostedServiceUrl(string url) {
      string regexString = @"[a-zA-Z0-9]+[a-zA-Z0-9-]{0,61}[a-zA-Z0-9]+";  //TODO
      Regex regex = new Regex(regexString);
      if (regex.IsMatch(url)) {
        return true;
      } else {
        return false;
      }
    }

    private bool isValidStorageContainerName(string name) {
      string regexString = @"[a-z0-9]+[a-z0-9-]{0,61}[a-z0-9]+";  //TODO
      Regex regex = new Regex(regexString);
      if (regex.IsMatch(name)) {
        return true;
      } else {
        return false;
      }
    }

    private void tbInstanceCount_Validating(object sender, CancelEventArgs e) {
      if (!isValidInstanceCount(tbInstanceCount.Text)) {
        errorProvider.SetError(tbInstanceCount, "Instance count must be a number greater than 0 and " +
                                                 "less or equal than the difference between maximum and available cores.");
        e.Cancel = true;
      } else {
        int instanceCount = Convert.ToInt32(tbInstanceCount.Text);
        Subscription subscription = (Subscription)cmbChooseSubscription.SelectedItem;
        if (instanceCount > subscription.MaxCoreCount - subscription.CurrentCoreCount) {
          errorProvider.SetError(tbInstanceCount, "Instance count must be less or equal than the difference between maximum and available cores.");
        } else {
          e.Cancel = true;
          errorProvider.SetError(tbInstanceCount, "");
        }
      }
    }

    private bool isValidInstanceCount(string text) {
      string regexString = @"[1-9]+[0-9-]*";  //TODO
      Regex regex = new Regex(regexString);
      if (regex.IsMatch(text)) {
        return true;
      } else {
        return false;
      }
    }

    private void tbBlobContainer_Validating(object sender, CancelEventArgs e) {
      string name = tbBlobContainer.Text.Trim();
      if (name == string.Empty) {
        errorProvider.SetError(tbBlobContainer, "A name is required for the blob container.");
        e.Cancel = true;
      } else if (!isValidStorageContainerName(name)) {
        errorProvider.SetError(tbBlobContainer, "A blob container name must be between 1 and 63 " +
                                              "characters long, and be composed of lowercase letters, numbers " +
                                              "and hyphens. Hyphens shouldn't be first or last letter.");
        e.Cancel = true;
      } else {
        errorProvider.SetError(tbBlobContainer, "");
      }
    }

    private void cmbChooseHostedService_Validating(object sender, CancelEventArgs e) {
      if ((!cbNewHostedService.Checked) && (cmbChooseHostedService.SelectedItem == null)) {
        errorProvider.SetError(cmbChooseHostedService, "A hoste service is required.");
        e.Cancel = true;
      } else {
        errorProvider.SetError(cmbChooseHostedService, "");
      }
    }

    private void cmbStorageServices_Validating(object sender, CancelEventArgs e) {
      if (cmbStorageServices.SelectedItem == null) {
        errorProvider.SetError(cmbStorageServices, "A storage service is required.");
        e.Cancel = true;
      } else {
        errorProvider.SetError(cmbStorageServices, "");
      }
    }

    private class CreateDeploymentData {
      public Subscription Subscription { get; set; }
      public bool CreateNewHosteService { get; set; }
      public bool UseRegion { get; set; }
      public string CertificateFilePath { get; set; }
      public string CertificatePassword { get; set; }
      public HostedService HostedService { get; set; }
      public int InstanceCount { get; set; }
      public StorageService StorageService { get; set; }
      public string BlobContainerName { get; set; }
      public bool CreateContainerIfNotExists { get; set; }
    }
  }
}
